package org.hmn.jrsd.demo;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;

import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.TableModel;

/*
 * 
 * Cree par Harry Karadimas a 11 sept. 08
 */
/**
 * A general-purpose table model, constructed from a list of java beans.
 * Each column is a property of the java bean. The list of columns is
 * constructed using reflection.
 */
public class JavaBeansTableModel
implements TableModel
{
  List<?> beans;
  HashSet<Integer> changeIndexes = new HashSet<Integer>();
  HashSet<String> forbiddenFields = new HashSet<String>();
  HashSet<String> readOnlyFields = new HashSet<String>();
  
  ArrayList<Method> getters = new ArrayList<Method>();
  HashMap<String, Method> settersByName = new HashMap<String, Method>();
  
  List<TableModelListener> listeners = new ArrayList<TableModelListener>();
  
  /**
   * Construct a table model for a collection of java beans. Here you should
   * declare fields that must be read-only (fields that correspond to
   * primary keys, mainly)
   * @param records
   * @param readOnlyFields
   */
  public JavaBeansTableModel(List<?> records, String... readOnlyFields)
  {
    declareForbidden("Class");
    setBeans(records);
    for (String field:readOnlyFields) declareReadOnly(field);
  }
  
  public void declareReadOnly(String fieldName)
  {
    readOnlyFields.add(fieldName);
  }
  
  public void declareForbidden(String fieldName)
  {
    forbiddenFields.add(fieldName);
  }
  
  public void addTableModelListener(TableModelListener l) {
    listeners.add(l);
  }

  public Class<?> getColumnClass(int columnIndex) {
    Method m = getters.get(columnIndex);
    //S ystem.out.println("Returning return type : "+m.getReturnType());
    Class<?> rt = m.getReturnType();
    if (rt.isPrimitive()) {
      if (rt == Integer.TYPE) return Integer.class;
    }
    return rt;
  }

  public int getColumnCount() { return getters.size(); }

  public String getColumnName(int columnIndex) {
    Method m = getters.get(columnIndex);
    return m.getName().substring(3);
  }

  public int getRowCount() {
    return beans.size();
  }

  public Object getValueAt(int rowIndex, int columnIndex) {
    Object o = beans.get(rowIndex);
    Method m = getters.get(columnIndex);
    try {
      return m.invoke(o);
    }
    catch (Exception ex) {
      throw new RuntimeException(ex);
    }
  }

  public boolean isCellEditable(int rowIndex, int columnIndex) {
    String fieldName = getColumnName(columnIndex);
    if (readOnlyFields.contains(fieldName)) return false;
    Method setter = findSetter(columnIndex);
    return setter != null;
  }

  private Method findSetter(int columnIndex)
  {
    Method getter = getters.get(columnIndex);
    String prop = getter.getName().substring(3);
    Method setter = settersByName.get(prop);
    return setter;
  }
  
  public void removeTableModelListener(TableModelListener l) {
    listeners.remove(l);
  }

  public void setValueAt(Object value, int rowIndex, int columnIndex) {
    Object bean = beans.get(rowIndex);
    Method setter = findSetter(columnIndex);
    if (setter == null) {
      throw new RuntimeException("No setter found for column "+columnIndex);
    }
    try {
      setter.invoke(bean, value);
      //record index of changed bean
      changeIndexes.add(rowIndex);
      //inform listeners of change
      TableModelEvent evt = new TableModelEvent(this, rowIndex);
      for (TableModelListener listener:listeners) listener.tableChanged(evt);
    }
    catch (Exception ex) {
      throw new RuntimeException(ex);
    }
  }

  private void fillGettersAndSetters()
  {
    for (Object bean:beans) {
      if (bean != null) {
        getters.clear();
        settersByName.clear();
        Class<?> bcl = bean.getClass();
        Method[] meths = bcl.getMethods();
        for (int i = 0; i < meths.length; i++) {
          Method m = meths[i];
          if (isMethodForForbiddenField(m)) continue;
          if (m.getName().length() <= 3) continue;
          if (!m.getName().startsWith("get")) continue;
          Class<?> rty = m.getReturnType();
          if (rty == Void.TYPE) continue;
          Class<?>[] ptypes = m.getParameterTypes();
          if (ptypes.length > 0) continue;
          getters.add(m);
        }
        for (int i = 0; i < meths.length; i++) {
          Method m = meths[i];
          if (isMethodForForbiddenField(m)) continue;
          if (m.getName().length() <= 3) continue;
          if (!m.getName().startsWith("set")) continue;
          Class<?> rty = m.getReturnType();
          if (rty != Void.TYPE) continue;
          Class<?>[] ptypes = m.getParameterTypes();
          if (ptypes.length == 0) continue;
          String propName = m.getName().substring(3);
          settersByName.put(propName, m);
        }
        break;
      }
    }//for
  }

  private boolean isMethodForForbiddenField(Method m)
  {
    String fieldName = m.getName().substring(3);
    return forbiddenFields.contains(fieldName);
  }
  
  public List<?> getBeans() {
    return beans;
  }

  public void setBeans(List<?> records) {
    this.beans = records;
    if (!records.isEmpty()) {
      //rebuild meta data
      getters.clear();
      settersByName.clear();
      changeIndexes.clear();
      fillGettersAndSetters();      
    }
    TableModelEvent evt = new TableModelEvent(this);
    for (TableModelListener listener:listeners) listener.tableChanged(evt);    
  }
  
  public List<?> getChangedBeans() 
  {
    List<Object> changedBeans = new ArrayList<Object>();
    for (int i:changeIndexes) changedBeans.add(beans.get(i));    
    return changedBeans;
  }
 
  public void clearChangeIndexes() { changeIndexes.clear(); }
  
  public Object getBean(int index) { return beans.get(index); }
  
}
