package org.hmn.jrsd;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.zip.CRC32;

import org.hmn.jrsd.rt.JrsdRtUtils;
import org.hmn.jrsd.rt.QColumn;

import antlr.RecognitionException;
import antlr.Token;

/*
 * 
 * Cree par Harry Karadimas a 30 juil. 2006
 */
/**
 * Descriptor for a column and its associated fields
 * @author karadimas
 */
public class ColumnDef
extends SqlElementDef
{
  TableDef parent;
  String name;
  ColumnTypeDef type;
  ArrayList<String> primaryKeys;
  ArrayList<FieldDecl> fieldDecls = new ArrayList<FieldDecl>();
  boolean key = false;
  List<String> directives;
  Boolean quoting;

  @Override
  public boolean isColumnDef() { return true; }

  /** constructor for primary key rule */
  public ColumnDef(
      TableDef parent, Token comment, ArrayList<String> pks, Token litSql)
  {
    this.parent = parent;
    setComment(mks(comment));
    this.primaryKeys = pks;
    setLiteralSql(mkl(litSql));
  }
  
  /** constructor for regular column declaration */
  public ColumnDef(
      TableDef parent, Token comment, String name, ColumnTypeDef type,
      Token litSql)
  {
    this.parent = parent;
    setComment(mks(comment));
    this.name = name;
    this.type = type;
    setLiteralSql(mkl(litSql));
  }
  
  /**
   * Adds the given field declaration. If it doesn't have a name, infer the
   * name from the column's name, converting it if necessary to upper or
   * lower case, depending on what option was given.
   * @param fd
   */
  public void add(FieldDecl fd)  
  {
    if (fd == null) throw new NullPointerException("fd est null");
    if (fd.getAlias() == null) {
      throw new RuntimeException("No alias defined");
    }
    fd.setColumnDef(this);
    if (fd.getFieldName() == null) {
      //sets the default field name, which is the name of the column.
      //The name will be corrected to be a correct java identifier.
      //The name is converted to upper case or to lower case if one the 
      //corresponding option was set.
      fd.setFieldName(Jrsd.condFieldCase(name));
    }
    MappingDecl md = parent.getMappingForAlias(fd.getAlias());
    if (md != null) {
      fieldDecls.add(fd);
      md.addFieldDecl(fd);
      return;
    }    
    else {
      MappingDecl gcl = 
        parent.getJrsdSpec().getGlobalColumnList(fd.getAlias());
      if (gcl != null) {
        gcl.addFieldDecl(fd);
        return;
      }
    }
    addError(new JrsdError("Alias '"+fd.getAlias()+"' unknown"));
    return;
  }

  /**
   * @param keyName
   * @return true if it is a primary key definition and keyName is in the 
   *   primaryKeys list
   */
  public boolean containsKey(String keyName)
  {
    return primaryKeys.contains(keyName);
  }
  
  public List<FieldDecl> selectFieldDecls(String aliasName)
  {
    ArrayList<FieldDecl> rl = new ArrayList<FieldDecl>();
    for (Iterator<FieldDecl> iter = fieldDecls.iterator(); iter.hasNext();) {
      FieldDecl fd = iter.next();
      if (aliasName.equals(fd.getParent().getAlias())) rl.add(fd);
    }
    return rl;
  }
  
  public FieldDecl[] getFieldDecls() {
    return (FieldDecl[]) fieldDecls.toArray(new FieldDecl[fieldDecls.size()]);
  }
  
  public List<String> getPrimaryKeys()
  {
    return primaryKeys;
  }
  
  public boolean isPrimaryKeyDecl() { return primaryKeys != null; }
    
  public TableDef getTableDef()
  {
    return parent;
  }
  
  public void setTableDef(TableDef tableDef)
  {
    this.parent = tableDef;
  }
  
  public String getName()
  {
    return name;
  }
  public String getQinName()
  {
    if (quoting != null) {
      if (quoting.booleanValue()) return "\"" + name + "\"";
      else return name;
    }
    return QColumn.quoteIfNecessary(name);
  }
  public void setName(String name)
  {
    this.name = name;
  }
  public TableDef getParent()
  {
    return parent;
  }
  public void setParent(TableDef parent)
  {
    this.parent = parent;
  }
  public ColumnTypeDef getType()
  {
    return type;
  }
  public void setType(ColumnTypeDef type)
  {
    this.type = type;
  }
  public boolean isEmpty(String str)
  {
    return str == null || str.trim().equals("");
  }
  public void addError(JrsdError e) { parent.addError(e); }
  
  public boolean isKey()
  {
    return key;
  }
  
  public void setKey(boolean key)
  {
    this.key = key;
  }
  
  public void update(CRC32 crc)
  {
    if (isPrimaryKeyDecl()) {
      for (Iterator<String> iter = primaryKeys.iterator(); iter.hasNext();) {
        String pkName = iter.next();
        JrsdRtUtils.update(crc, pkName);
      }
    }
    else {
      JrsdRtUtils.update(crc, parent.getName());
      JrsdRtUtils.update(crc, name);
      type.update(crc);      
    }
  }
  
  /**
   * Supported directives : quote, unquote 
   * @param directives
   * @throws RecognitionException
   */
  public void addDirectives(List<String> directives) 
  throws RecognitionException
  {
    for (Iterator<String> iter = directives.iterator(); iter.hasNext();) {
      String directive = iter.next();
      if (directive.equals("quote")) {
        quoting = Boolean.TRUE;
      }
      else if (directive.equals("unquote")) {
        quoting = Boolean.FALSE;
      }
      else {
        throw new RecognitionException("Unknown column directive '"+directive+"'");
      }
    }//for    
  }

  public Boolean getQuoting() {
    return quoting;
  }

  public void setQuoting(Boolean quoting) {
    this.quoting = quoting;
  }

}
