package org.hmn.jrsd;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;

import org.hmn.jrsd.rt.JrsdRtUtils;

import antlr.RecognitionException;
import antlr.Token;

/*
 * 
 * Cree par Harry Karadimas a 31 juil. 2006
 */
/**
 * Field declaration.
 * There are two cases : only a field type has been declared, or a field type
 * and a base type have been declared.
 * @author karadimas
 */
public class FieldDecl
implements Comparable<FieldDecl>
{
  MappingDecl parent;
  String alias;
  String fieldName;
  String getterSetterName;
  ColumnDef cdef;
  String explicitSFieldType;
  List<String> sormFlags = new ArrayList<String>();
  ArrayList<FieldDecl> subfields = new ArrayList<FieldDecl>();
  FieldDecl owner;
  String sormBaseType; //base type, e.g. "Boolean" when SFieldBoolean abc = new SFieldBooleanCharTF(...)
  boolean accessorGenerationEnabled = true;
  boolean sequenceGenerator = false;
  boolean insertGenerator = false;
  boolean selectmaxGenerator = false;
  boolean unqueried = false;

  //Allowed flags. Copied from simpleorm v.3 source
  private static final String[] ALLOWED_SFIELD_FLAGS_ARRAY = {
      "SPRIMARY_KEY",
      "SMANDATORY",
      "SNOT_OPTIMISTIC_LOCKED", 
      "SDESCRIPTIVE",
      "SUNQUERIED"
  };
  private static final HashSet<String> ALLOWED_SFIELD_FLAGS =
    new HashSet<String>(Arrays.asList(ALLOWED_SFIELD_FLAGS_ARRAY));

  //types whose getter/setter name is identical to the type
  private static final String[] IDENTICAL_ACCESSOR_TYPES_ARRAY = {
    "BigDecimal", "Bytes", "int", "Date", "Enum", "String", "Time", "Timestamp"
  };
  
  private static final HashSet<String> IDENTICAL_ACCESSOR_TYPES =
    new HashSet<String>(Arrays.asList(IDENTICAL_ACCESSOR_TYPES_ARRAY));
  
  //types whose return type is identical to the type in the SFieldXxx
  private static final String[] IDENTICAL_RETURN_TYPES_ARRAY = {
    "BigDecimal", "Bytes", "Date", "Enum", "String", "Time", "Timestamp"
  };
  
  private static final HashSet<String> IDENTICAL_RETURN_TYPES =
    new HashSet<String>(Arrays.asList(IDENTICAL_RETURN_TYPES_ARRAY));
  
  public boolean isColumnDef() { return false; }
  
  public FieldDecl(ColumnDef cd)
  {    
    if (cd == null) throw new NullPointerException("cd est null");
    this.cdef = cd;
  }
  
  public FieldDecl(Token alias, Token fn) 
  {
    this.alias = alias == null ? null : alias.getText();
    this.fieldName = (fn==null) ? null : fn.getText() ;
  }
  
  /**
   * Add a subfield. A subfield is used to declare an attribute in a
   * transport object.
   * @param fd
   */
  public void addSubfield(FieldDecl fd)
  {
    if (cdef == null) throw new NullPointerException("cdef est null");
    MappingDecl md = cdef.getParent().getMappingForAlias(fd.getAlias());
    if (md == null) {
      addError(new JrsdError("Unknown alias '"+fd.getAlias()+"'"));
    }    
    subfields.add(fd);
    fd.owner = this;
  }
  
  public List<FieldDecl> getSubfields() { return subfields; }
  
  /**
   * Get the field's name. In case of a subfield, if no name was given for
   * the subfield, take the name of the owner field.
   * @return
   */
  public String getFieldName()
  {
    if (isEmpty(fieldName) && owner != null) {
      return owner.getFieldName();
    }    
    return fieldName; 
  }
    
  /**
   * get the field name, with the prefix added. If the prefix is non-empty and
   * does not end with "_", capitalize first letter of field name.
   * For example with field "foo" and prefix "field_", the prefixed name will
   * be "field_foo", and with prefix "f" it will be "fFoo"
   * @return
   */
  public String getPrefixedFieldName()
  {
    String prefix = getParent().getJrsdSpec().getFieldPrefix();
    String fname = getFieldName();
    if (isEmpty(prefix)) return fname;
    if (!prefix.endsWith("_")) fname = JrsdRtUtils.capitalizeFirst(fname);
    return prefix + fname;
  }
  
  /**
   * Set the name of the field, and also the name of the accessor if 
   * it is not defined (with the first letter converted to upper case).
   * @param fn
   */
  public void setFieldName(String fn)
  {
    if (fn == null) throw new NullPointerException("fieldName is null");
    String jfn = JrsdUtils.correctForJava(fn);
    fieldName = jfn;
    if (getterSetterName == null) {
      getterSetterName = Character.toUpperCase(jfn.charAt(0))+jfn.substring(1);
    }
  }
  
  public void setFieldName(Token t) { 
    setFieldName(t == null ? null : t.getText()); 
  }
  
  public MappingDecl getParent()
  {
    return parent;
  }
  
  public void setParent(MappingDecl parent)
  {
    this.parent = parent;
  }
  
  public ColumnDef getColumnDef()
  {
    return cdef;
  }
  
  public void setColumnDef(ColumnDef columnDef)
  {
    this.cdef = columnDef;
  }
  
  public void addError(JrsdError e) { cdef.addError(e); }
  
  public String getAlias()
  {
    return alias;
  }
  
  public void setAlias(String alias)
  throws RecognitionException
  {
    parent = cdef.getParent().getMappingForAlias(alias);
    if (parent == null) {
      MappingDecl gmd =
        cdef.getParent().getJrsdSpec().getGlobalColumnList(alias);
      if (gmd == null)
        throw new RuntimeException("Unknown alias '"+alias+"'");
    }
    this.alias = alias;
  }
  
  public void setAlias(Token t)
  throws RecognitionException
  {
    setAlias(t == null ? null : t.getText());
  }
  
  public String getExplicitSFieldTypeName()
  {
    return explicitSFieldType;
  }
  
  public boolean isSormReference() {
    if (explicitSFieldType == null) return false;
    return explicitSFieldType.startsWith("REF#");
  }
  
  public String getSormReferenceName() { return explicitSFieldType.substring(4); }
  
  /**
   * Explicite name of a java Class to use for this field. Normally null,
   * default values are used using what was declared in sqlType.
   * @see #getInferredSFieldType()
   * @param explicitSFieldType
   */
  public void setExplicitSFieldTypeName(String st)
  {
    explicitSFieldType = st;
    //Replace correct case for some types    
    //if ("BigDecimal".equalsIgnoreCase(sormType)) {
    //  sormType = "BigDecimal";
    //}    
  }
  
  public void setExplicitSFieldTypeName(Token t)
  {
    setExplicitSFieldTypeName(t == null ? null : t.getText());
  }
  
  /**
   * @return The type name that was declared in the column, in upper case, for
   *   reliable comparison.
   */
  private String getColTypeUc()
  {
    if (cdef == null) return null;
    String colType = cdef.getType().getTypeName().toUpperCase();
    return colType;
  }
  
  public String getInferredSFieldType()
  {
    if (explicitSFieldType != null) return explicitSFieldType;
    String colType = getColTypeUc();
    
    //primitive types
    if (colType.equals("PRIMITIVE_BOOLEANBIT")) return "BooleanBit";
    if (colType.equals("PRIMITIVE_BOOLEANCHAR")) return "BooleanChar";    
    if (colType.equals("INT")) return "Integer";
    if (colType.equals("PRIMITIVE_LONG")) return "Long";
    if (colType.equals("PRIMITIVE_DOUBLE")) return "Double";
    
    //primitive types, as objects
    if (colType.equals("BOOLEANBIT")) return "BooleanBit";
    if (colType.equals("BOOLEANCHAR")) return "BooleanChar";    
    if (colType.equals("INTEGER")) return "Integer";
    if (colType.equals("LONG")) return "Long";
    if (colType.equals("DOUBLE")) return "Double";
    
    //regular objects
    if (colType.equals("VARCHAR")) return "String";
    if (colType.equals("CHAR")) return "String";
    if (colType.equals("NUMBER")) {
      if (isEmpty(getColumnDef().getType().getScale())) return "Long";
      else return "BigDecimal";
    }
    if (colType.equals("BIGDECIMAL")) return "BigDecimal";
    if (colType.equals("DATE")) return "Date";
    if (colType.equals("TIMESTAMP")) return "Timestamp";
    if (colType.equals("TIME")) return "Time";
    if (colType.equals("BYTES")) return "Bytes";
    if (colType.equals("BLOB")) return "Blob"; //SFieldBlob may not (yet) be what we want
    if (colType.equals("CLOB")) return "Clob"; //SFieldClob may not (yet) be what we want
    if (colType.equals("OTHER")) return cdef.getType().getParam1();
    return "??"+colType;    
  }
  
  /**
   * 
   * @return The Xxx for SField.getXxx, where getInt, getBoolean, getDouble are
   *   replaced with getObject
   */
  public String getSFieldGetterNameForObject()
  {
    String gn = getSFieldGetterName();
    if (gn.equals("Boolean")) return "Object";
    if (gn.equals("Int")) return "Object";
    if (gn.equals("Long")) return "Object";
    if (gn.equals("Double")) return "Object";
    return gn;
  }
  
  /**
   * Get the best name for SField.getXxx in Simpleorm. Beware of
   * getInt, getBoolean, getLong, getDouble, as they return only the primitive values,
   * therefore masking the NULL values.
   * @return
   */
  public String getSFieldGetterName()
  {
    String t = getInferredSFieldType();    
    if (IDENTICAL_ACCESSOR_TYPES.contains(t)) return t;
    t = getColTypeUc();
    if (t.equals("INT")) return "Int";
    if (t.equals("PRIMITIVE_BOOLEAN")) return "Boolean";
    if (t.equals("PRIMITIVE_BOOLEANCHAR")) return "Boolean";
    if (t.equals("PRIMITIVE_BOOLEANBIT")) return "Boolean";
    if (t.equals("PRIMITIVE_DOUBLE")) return "Double";
    if (t.equals("PRIMITIVE_LONG")) return "Long";
    return "Object";
  }
  
  public String getSFieldSetterName()
  {
    return getSFieldGetterName();
  }
  
  public String getSFieldSetterNameForObject()
  {
    return getSFieldGetterNameForObject();
  }
  
  /**
   * Sometimes a SFieldZZZ does not return exactly ZZZ. This is the case
   * for example for SFieldFiniteString, which returns a String, or SFieldBytes,
   * which returns a byte[]. This method returns the java type of the
   * returned value.
   * @return
   */
  public String getSormReturnedType()
  {
    //First look in the base type, which has precedence
    String typ = getBaseSFieldType();
    //If no base type defined, look for an explicit type name
    if (typ == null) typ = getExplicitSFieldTypeName();
    if (typ != null) return getJavaTypeForSormType(typ);    
    
    //get the column type and look for special cases (the primitive types)
    typ = getColTypeUc();
    
    if (typ == null) typ = "Object";
    //look for special primitive
    if (typ.equals("PRIMITIVE_BOOLEANCHAR")) return "boolean";
    if (typ.equals("PRIMITIVE_BOOLEANBIT")) return "boolean";
    if (typ.equals("INT")) return "int";
    if (typ.equals("PRIMITIVE_LONG")) return "long";
    if (typ.equals("PRIMITIVE_DOUBLE")) return "double";      
    
    //get the return type from the column declaration
    typ = getInferredSFieldType();
    
    return getJavaTypeForSormType(typ);
  }
  
  String getJavaTypeForSormType(String typ)
  {
    if (typ.equals("FiniteString")) return "String";
    if (typ.equals("Bytes")) return "byte[]";
    if (typ.equals("Date")) return "java.sql.Date";
    if (typ.equals("BooleanChar")) return "Boolean";
    if (typ.equals("BooleanBit")) return "Boolean";
    return typ;
  }
  
  /**
   * Same as getSormReturnedType but translates primitive types to their
   * object equivalent
   * @return
   */
  public String getSormReturnedObjectType()
  {
    String t = getSormReturnedType();
    if (t.equals("boolean")) return "Boolean";
    if (t.equals("int")) return "Integer";
    if (t.equals("long")) return "Long";
    if (t.equals("double")) return "Double";
    return t;
  }
  
  /**
   * SFieldX where X can be : <br>
   * BigDecimal, Boolean, Date, Double, Integer, Long, Object, String, Time, Timestamp
   * <p>
   * Corresponding primitive types are : 
   * boolean, double, int, long
   * <p>
   * Note that the 8 primitive java types are : 
   * boolean, byte, short, int, long, float, double, char 
   * @return True if one of the explicit or implicit types is primitive
   */
  public boolean isPrimitiveSormType()
  {
    //look if a type was declared explicitly as int, boolean, long, double
    String typ = getBaseOrNormalSFieldType();
    if (isPrimitiveType(getSFieldType())) return true;
    //look if a column was declared as "primitive_boolean", "primitive_long",
    //"primitive_double", or "int"
    typ = getColTypeUc();
    return 
        typ.equals("PRIMITIVE_BOOLEAN")
      | typ.equals("PRIMITIVE_BOOLEANCHAR")
      | typ.equals("PRIMITIVE_BOOLEANBIT")
      | typ.equals("PRIMITIVE_LONG")
      | typ.equals("PRIMITIVE_DOUBLE")
      | typ.equals("INT")
    ;
  }
  
  /**
   * Is the java type a primitive type
   * @param typ
   * @return
   */
  public boolean isPrimitiveType(String typ)
  {
    return typ != null && ( 
        typ.equals("boolean") || 
        typ.equals("int") || 
        typ.equals("long") ||         
        typ.equals("double")
        );
  }
  
  public boolean isSormTypeBoolean()
  {
    if (sormBaseType == null) return "boolean".equalsIgnoreCase(explicitSFieldType);
    return "boolean".equalsIgnoreCase(sormBaseType);
  }
  
  /** is the field of type CLOB or BLOB */
  public boolean isLob()
  {
    String bt = getBaseOrNormalSFieldType();
    return bt.equalsIgnoreCase("clob") || bt.equalsIgnoreCase("blob");
  }
  
  public String getSormTypeForAccessor()
  {
    String bt = getBaseOrNormalSFieldType();
    if (bt.equals("FiniteString")) bt = "String";
    return JrsdRtUtils.capitalizeFirst(bt);    
  }
  
  /**
   * Get the Xxx for the sorm SFieldXxx. This is usually the same
   * string as the type, except for some reserved primitive types.
   * @return
   */
  public String getSFieldType()
  {
    if (explicitSFieldType != null) return explicitSFieldType;
    //no type explicitly declared, infer SField type from column declaration
    String t2 = getInferredSFieldType();    
    return t2;
  }
    
  public String getBaseOrNormalSFieldType()
  {
    String bt = getBaseSFieldType();
    if (bt != null) return bt;
    return getSFieldType();
  }
  
  /**
   * Get the equivalent Object type for the primitive type
   * @param typ
   * @return
   */
  public String getObjectType(String typ)
  {
    if (typ.equals("boolean")) return "Boolean";
    if (typ.equals("int")) return "Integer";
    if (typ.equals("long")) return "Long";
    if (typ.equals("double")) return "Double";
    return typ;    
  }
  
  public void addDirectives(List<String> flags)
  throws RecognitionException
  {
    for (Iterator<String> iter = flags.iterator(); iter.hasNext();) {
      String directive = iter.next();
      if (ALLOWED_SFIELD_FLAGS.contains(directive)) {
        //sprimary_key, smandatory, snot_optimistic_locked, sdescriptive, sunqueried
        sormFlags.add(directive);
      }
      else if (directive.equals("noaccessors")) {
        accessorGenerationEnabled = false;
        //S ystem.out.println("Accessor generation disabled.");
      }
      else if (directive.equals("usesequencegen")) {
        setSequenceGenerator(true); 
        //sormFlags.add("SGENERATED_KEY.pvalue(new SGeneratorSequence(meta))");
      }
      else if (directive.equals("useinsertgen")) {
        setInsertGenerator(true);
        //sormFlags.add("SGENERATED_KEY.pvalue(new SGeneratorSequence(meta))");
      }
      else if (directive.equals("useselectmaxgen")) {
        setSelectmaxGenerator(true);
        //sormFlags.add("SGENERATED_KEY.pvalue(new SGeneratorSequence(meta))");
      }
      else {
        throw new RecognitionException("Unknown directive '"+directive+"'");
      }
    }//for    
  }
  
  
  void updateSormFlags()
  {
    if (cdef.isKey() && !sormFlags.contains("SPRIMARY_KEY")) {
      sormFlags.add("SPRIMARY_KEY");
    }
  }
  
  public boolean getContainsSormFlags() 
  {
    updateSormFlags();
    return !sormFlags.isEmpty(); 
  }
  
  public List<String> getSormFlags()
  {
    updateSormFlags();
    return sormFlags;
  }
  
  /**
   * Get the owner of this subfield
   * @return
   */
  public FieldDecl getOwner()
  {
    return owner;
  }
  
  public void setOwner(FieldDecl owner)
  {
    this.owner = owner;
  }
  
  public String getStringFieldSize()
  {
    String DEFLEN = "999999";
    if (!"String".equals(getInferredSFieldType())) return null;
    if (cdef == null) return DEFLEN;
    ColumnTypeDef ctd = cdef.getType();
    if (ctd == null) return DEFLEN;
    if (ctd.getPrecision() == null) return DEFLEN;
    return ctd.getPrecision();
  }

  /**
   * @see java.lang.Comparable#compareTo(java.lang.Object)
   */
  public int compareTo(FieldDecl q)
  {
    if (q == null) return -1;
    String fnp = String.valueOf(fieldName);
    String fnq = String.valueOf(q.fieldName);
    return fnq.compareTo(fnp);
  }
  /**
   * Le nom de l'accesseur est le nom utilis pour les mthodes getXXX, setXXX.
   * @return
   */
  public String getGetterSetterName() {
    return getterSetterName;
  }
  public void setGetterSetterName(String getterSetterName) {
    this.getterSetterName = getterSetterName;
  }
  
  /**
   * Get the simple orm base type. This is usually the same as the sorm type,
   * except when the "::" construct was used to declare a base class. This is
   * mostly used in cases where the constructor class is not the general
   * class (for example Boolean and BooleanCharTF, where the generated code
   * will be SFieldBoolean abc = new SFieldBooleanCharTF(...)).
   * @return
   */
  public String getBaseSFieldType() {
    return sormBaseType;    
  }
     
  public void setSormBaseType(String sormBaseType) {
    this.sormBaseType = sormBaseType;
  }
  
  public boolean isAccessorGenerationEnabled() {
    return accessorGenerationEnabled;
  }
  public void setAccessorGenerationEnabled(boolean accessorGenerationEnabled) {
    this.accessorGenerationEnabled = accessorGenerationEnabled;
  }
  
  public boolean isPrecisionNeeded()
  {
    String bt = getSFieldType();
    String colTypeUc = getColTypeUc();   
    //S ystem.out.println("bt:"+bt);
    return "BigDecimal".equals(bt) 
        || "String".equals(bt)
        || "FiniteString".equals(bt)
        || "BooleanChar".equals(bt)
        || "Bytes".equals(bt)
        || "Blob".equals(bt)
        || "Clob".equals(bt)
        ;
  }
  
  public boolean isScaleNeeded()
  {
    String bt = getBaseOrNormalSFieldType();
    if ("BigDecimal".equals(bt)) return true;
    if ("BooleanChar".equals(bt)) return true;
    //special case for OTHER where user can add something in "scale" (except for "Object")
    if (getColTypeUc().equals("OTHER")) {
      return cdef.getType().getScale() != null && !bt.equals("Object");
    }
    return false;
  }
   
  public String getPrecision()
  {
    String prec = getColumnDef().getType().getPrecision();
    if (isEmpty(prec)) prec = "999999";
    String bt = getSFieldType();
    //S ystem.out.println("bt="+bt);
    if ("BooleanChar".equals(bt)) return '"'+prec+'"';
    return prec;
  }
  
  public String getScale()
  {
    String sc = getColumnDef().getType().getScale();
    if (isEmpty(sc)) sc = "0";
    if ("BooleanChar".equals(getBaseOrNormalSFieldType())) return '"'+sc+'"';
    return sc;
  }
  
  public boolean isEmpty(String str) {
    return str == null || str.trim().equals("");
  }
  
  /**
   * @see java.lang.Object#toString()
   */
  public String toString() {
    StringBuffer sb = new StringBuffer();
    sb.append("FieldDecl[alias="+alias+"]");
    return sb.toString();
  }
  
  /**
   * Look for a "foreign key" column, i.e. a field that has been
   * declared for this column, and which is not a reference.
   * @return
   */
  public String getForeignKeyColumn()
  {
    List<FieldDecl> fds = getParent().getFieldDeclarations();
    for (Iterator<FieldDecl> iter = fds.iterator(); iter.hasNext();) {
      FieldDecl fd = (FieldDecl) iter.next();
      if (cdef.equals(fd.cdef) & !fd.isSormReference()) {
        return fd.getFieldName();
      }
    }//for
    return null;
  }

  public boolean isSequenceGenerator() {
    return sequenceGenerator;
  }

  public void setSequenceGenerator(boolean sequenceGenerator) {
    this.sequenceGenerator = sequenceGenerator;
  }
  
  public boolean isInsertGenerator() {
    return insertGenerator;
  }

  public void setInsertGenerator(boolean insertGenerator) {
    this.insertGenerator = insertGenerator;
  }
  
  public boolean isSelectmaxGenerator() {
    return selectmaxGenerator;
  }

  public void setSelectmaxGenerator(boolean selectmaxGenerator) {
    this.selectmaxGenerator = selectmaxGenerator;
  }
  
  //-------------
  //--- Methods specialized for code generation
  //-------------
  
  public String getCgReturnType()
  {
    return getSormReturnedType();
  }
  
  public String getCgReturnTypeAsObject()
  {
    return getObjectType(getSormReturnedType());
  }
  
  public String getCgJavaType() { return getCgReturnType(); }
  
  public String getCgJavaTypeAsObject() { return getCgReturnTypeAsObject(); }
  
  /**
   * @return The base type; if no base type was explicitly declared, return
   *   the regular type.
   */
  public String getCgBaseType()
  {
    String bt = getBaseOrNormalSFieldType();
    //special case #1 : subtitute Blob for byte[] (kludge for SFieldBlob support)
    if ("byte[]".equals(bt)) bt = "Blob";
    return bt;
  }
  
  /**
   * @return Normal type only; ignore the base type.
   */
  public String getCgFieldType()
  {
    return getSFieldType();
  }
  
  public boolean isQuotingNecessary()
  {
    Boolean qn = getColumnDef().getQuoting();
    if (qn != null) return qn.booleanValue();
    String colName = getColumnDef().getName();
    return Jrsd.isQuotingNecessary(colName);
  }

  public boolean isUnqueried() {
    return unqueried;
  }

  public void setUnqueried(boolean unqueried) {
    this.unqueried = unqueried;
  }

  /**
   * To find the name of the field that is referenced, we use the first
   * declared field for the column.
   * @return
   */
  public String getPrefixedForeignKeyNameForReference()
  {
    return getColumnDef().getFieldDecls()[0].getPrefixedFieldName();
  }
  
}
