package org.hmn.jrsd.rt;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.sql.Blob;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.zip.CRC32;

import simpleorm.dataset.SFieldScalar;
import simpleorm.dataset.SRecordInstance;
import simpleorm.dataset.SRecordMeta;


/*
 * 
 * Cree par Harry Karadimas a 16 aot 2006
 */
/**
 * Misc static methods used by jrsd.
 * @author karadimas
 */
public class JrsdRtUtils
{

  /**
   * Update the CRC with the ISO-8859-1 bytes of the string.
   * @param crc
   * @param str
   */
  static public void update(CRC32 crc, String str)
  {
    byte[] strBytes;
    if (str == null) return;
    try {
      strBytes = str.getBytes("ISO-8859-1");
    }
    catch (UnsupportedEncodingException ex) {
      throw new RuntimeException("Critical : encoding ISO-8859-1 not found");
    }
    crc.update(strBytes);
  }  
  
  /**
   * Simple read of a text file. This uses the platform's encoding.
   * @param relPath
   * @return
   * @throws IOException
   */
  static public String loadFile(String relPath)
  throws IOException
  {
    File f = new File(relPath);
    FileReader fr = new FileReader(f);
    StringBuffer sb = new StringBuffer();
    int ch = 0;
    while ((ch = fr.read()) >= 0) sb.append((char)ch);
    return sb.toString();
  }
  
  /**
   * Sets the first letter of the string to uppercase
   * @param str
   * @return
   */
  public static String capitalizeFirst(String str)
  {
    if (str == null) return null;
    StringBuffer sb = new StringBuffer();
    if (str.length() == 0) return "";
    sb.append(Character.toUpperCase(str.charAt(0)));
    if (str.length() > 1) sb.append(str.substring(1));
    return sb.toString();
  }


  /**
   * Replace special characters with entities :
   * <UL>
   *   <LI>&amp; -&gt; &amp;amp;
   *   <LI>&lt; -&gt; &amp;lt;
   *   <LI>&gt; -&gt; &amp;gt;
   *   <LI>&quot; -&gt; &amp;quot;
   *   <LI>&#39; -&gt; &amp;#39;
   * </UL>
   * Use for regular text you want to display as it is in HTML.   
   */
  public static String encodeWithEntities(String str)
  {
    return JrsdRtUtils.encodeWithEntities(str, ' ');
  }

  /**
   * Same as encodeWithEntities(String str) but specifies a char to leave
   * untouched (only '\'' or '"' are possible here). This is useful for example
   * in javascript constants. Also optimized to allocate a StringBuffer only if
   * necessary.
   * @param str
   * @param preserve
   * @return
   */
  public static String encodeWithEntities(String str, char preserve)
  {
    if (str == null) return null;
    int len = str.length();
    StringBuffer sb = null; 
    for (int i = 0; i < len; i++) {
      char c = str.charAt(i);
      switch (c) {
      case '&':
        if (sb == null) sb = new StringBuffer(str.substring(i));
        sb.append("&amp;");
        break;
      case '<':
        if (sb == null) sb = new StringBuffer(str.substring(i));
        sb.append("&lt;");
        break;
      case '>':
        if (sb == null) sb = new StringBuffer(str.substring(i));
        sb.append("&gt;");
        break;
      case '"':
        if (sb == null) sb = new StringBuffer(str.substring(i));
        if (c == preserve)
          sb.append(c);
        else
          sb.append("&quot;");
        break;
      case '\'':
        if (sb == null) sb = new StringBuffer(str.substring(i));
        if (c == preserve)
          sb.append(c);
        else
          sb.append("&#39;");
        break;
      default:
        if (sb != null) sb.append(c);
      }
    }
    if (sb == null) return str;
    return sb.toString();
  }
    
  
  public static Blob lockRowAndSelectBlob(
      Connection cxn, String blobCol, SRecordInstance rec)
  throws IOException, SQLException
  {
    //Before we had :
    //SRecordMeta<? extends SRecordInstance> rmeta = rec.getMeta();
    //But with 3.3r845 rec.getMeta() returns SRecordMeta<?>
    SRecordMeta<?> rmeta = rec.getMeta();
    SFieldScalar[] pks = rmeta.getPrimaryKeys();
    String[] pkNames = new String[pks.length];
    Object[] pkValues = new String[pks.length];
    for (int i = 0; i < pks.length; i++) {
      pkNames[i] = pks[i].getColumnName();
      pkValues[i] = rec.getObject(pks[i]);
    }
    return JrsdRtUtils.lockRowAndSelectBlob(
        cxn, rmeta.getTableName(), blobCol, pkNames, pkValues);
  }

  /**
   * Lock row explicitly. Useful in some obscure cases, for records that can't 
   * be used in in simpleorm, because their composite primary key can contain a
   * null value (nulls are not supported for primary key values in simple orm,
   * but are in some exotic databases, I can't even remember which ones).
   * Try not to use this method and keep your primary key values not null.
   * Note that your database must support the SELECT ... FOR UPDATE sql syntax.
   * @param table
   * @param pkNames
   * @param pkVals
   * @throws IOException
   */
  public static Blob lockRowAndSelectBlob(
      Connection cxn, String table, String blobCol, String[] pkNames, Object[] pkVals)
  throws IOException, SQLException
  {
    StringBuffer sb = new StringBuffer();
    sb.append("SELECT \"" + blobCol + "\" FROM ");
    sb.append(QColumn.quoteIfNecessary(table));
    AndExpr clauses = new AndExpr();
    clauses.addAllEqClauses(pkNames, pkVals);
    sb.append(clauses.getText());
    sb.append(" FOR UPDATE");
    String sql = sb.toString();
    PreparedStatement ps = cxn.prepareStatement(sql);
    try {
      clauses.setParameters(1, ps);
      ResultSet rs = ps.executeQuery();
      Blob blb = null;
      if (rs.next()) blb = rs.getBlob(1);      
      return blb;
    }
    finally {
      ps.close();
    }
  }
  
  /**
   * Construct a list, with each element enclosed in quoteChar(s) if quoteChar
   * is != 0, and each element separated from the others by the eltSep string.
   * @param strs
   * @param quoteChar
   * @param eltSep
   * @return
   */
  public static String makeList(String[] strs, char quoteChar, String eltSep)
  {
    StringBuffer sb = new StringBuffer();
    for (int i = 0; i < strs.length; i++) {
      if (i > 0) sb.append(eltSep);
      if (quoteChar != 0) sb.append(quoteChar);
      sb.append(strs[i]);
      if (quoteChar != 0) sb.append(quoteChar);
    }
    return sb.toString();
  }
  
}
