package org.hmn.jrsd.rt;

import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import org.apache.log4j.Logger;

import simpleorm.dataset.SFieldMeta;
import simpleorm.dataset.SFieldReference;
import simpleorm.dataset.SFieldString;
import simpleorm.dataset.SRecordInstance;

/*
 * 
 * Cree par Harry Karadimas a 9 aot 2006
 */
/**
 * TransportMappings stores associations between simpleorm fields and a
 * user-defined name (which can different from the name of the mapping). For
 * each "xxx" name, the first "setXxx" method that
 * is found will be used (therefore avoid having several setXxx methods).
 * @author karadimas
 */
public class TransportMappings
implements Serializable
{
  private static final long serialVersionUID = 1597658622719314899L;

  static Logger lg = Logger.getLogger(TransportMappings.class);
  
  static final Class<Number> numberClass = Number.class;
    
  static public boolean debugValueTransfer = false;
  
  HashMap<Class<?>, ArrayList<TransportFieldMapping>> mappingsByClass = 
    new HashMap<Class<?>, ArrayList<TransportFieldMapping>>();
  
  /**
   * Init this Mappings translation object, with an object array that has
   * following elements :
   * Object class, sfield, att_name, [, sfield, att_name ...]
   * [, Object class, sfield, att_name, [, sfield, att_name ...] ...]
   * @param mappings
   */
  public TransportMappings(Object[] mappings) 
  {
    Class<?> curClass = null;
    SFieldMeta curMeta = null;    
    for (int i = 0; i < mappings.length; i++) {
      Object obj = mappings[i];
      if (obj instanceof Class) {
        curClass = (Class<?>) obj;
      }
      else if (obj instanceof SFieldMeta) {
        curMeta = (SFieldMeta) obj;
      }
      else {
        String name = (String) obj;
        ArrayList<TransportFieldMapping> mappingList = 
          (ArrayList<TransportFieldMapping>) mappingsByClass.get(curClass);
        if (mappingList == null) {
          mappingList = new ArrayList<TransportFieldMapping>();
          mappingsByClass.put(curClass, mappingList);
        }
        TransportFieldMapping mping = new TransportFieldMapping(curMeta, name);
        mappingList.add(mping);
      }
    }//for
  }
  
  /**
   * Write values from record to transport object.
   * @param rec
   * @param transportObject
   */
  public void writeValuesFromTo(SRecordInstance rec, Object transportObject)  
  {
    Class<?> tObjClass = transportObject.getClass();
    ArrayList<TransportFieldMapping> mappings = mappingsByClass.get(tObjClass);
    if (mappings == null) {
      throw new ValueTransferException("No transport field mappings found " +
          "for object "+transportObject);
    }
    Iterator<TransportFieldMapping> iter;
    for (iter = mappings.iterator(); iter.hasNext(); ) {
      TransportFieldMapping mping = iter.next();
      SFieldMeta fm = mping.fm;
      String attName = mping.attName;
      //get values as objects. "Invalid" field values (they are invalid because
      //they haven't been queries by simpleorm) are interpreted as null
      Object val = rec.isValid(fm) ? rec.getObject(fm) : null;
      if (val == null) {
        ReflectionHelpers.setNullToAttribute(transportObject, attName);
      }
      else {
        ReflectionHelpers.setValueToAttribute(transportObject, attName, val);
      }
    }// for
  }
    
  /**
   * Write values from transport object to record. Detects and does NOT write
   * into : primary key fields, references.
   * (reference foreign keys can be now written to, as with simpleorm v.3 we
   * can not detect them)
   * @param transportObject
   * @param rec
   */
  public void writeValuesFromTo(Object transportObject, SRecordInstance rec)
  {
    Class<?> tObjClass = transportObject.getClass();
    ArrayList<TransportFieldMapping> mappings = mappingsByClass.get(tObjClass);
    if (mappings == null) {
      throw new ValueTransferException("No transport field mappings found " +
          "for object "+transportObject+" of class "+tObjClass);
    }
    Iterator<TransportFieldMapping> iter;
    for (iter = mappings.iterator(); iter.hasNext(); ) {
      TransportFieldMapping mping = iter.next();
      SFieldMeta fm = mping.fm;
      String attName = mping.getAttributeName();
      if (SormUtils.isPartOfPrimaryKey(fm)) continue;
      if (fm instanceof SFieldReference<?>) continue;      
      //if (SFieldAccessor.isReferenceKey(fm)) continue;
      Object val = getValueFromAttribute(transportObject, attName);
      if (val == null) setNullToField(rec, fm);
      else setValueToField(rec, fm, val);
    }// for    
  }
  
  private Object getValueFromAttribute(Object obj, String propName)
  {
    Class<?> objCls = obj.getClass();
    String getMethName = "get"+ReflectionHelpers.capitalizeFirst(propName);
    try {      
      Method m = objCls.getMethod(getMethName);      
      if (debugValueTransfer) { 
        lg.debug("Appel de getValueFromAttribute pour "+getMethName);
      }
      Object r = m.invoke(obj);
      return r;
    }
    catch (NoSuchMethodException nsmex) {
      String isMethName = "is"+ReflectionHelpers.capitalizeFirst(propName);
      try {      
        Method m = objCls.getMethod(isMethName);      
        if (debugValueTransfer) { 
          lg.debug("Appel de isValueFromAttribute pour "+isMethName);
        }
        Object r = m.invoke(obj);
        return r;
      }
      catch (NoSuchMethodException nsmex2) {
        String msg = "No " + getMethName + " or " + 
                     isMethName + " method found";
        throw new ValueTransferException(msg);
      }
      catch (Exception e) {
        lg.error("(prop["+propName+"])", e);
        throw new ValueTransferException(e);
      }
    }
    catch (Exception e) {
      lg.error("(prop["+propName+"])", e);
      throw new ValueTransferException(e);
    }    
  }
  
  private void setNullToField(SRecordInstance rec, SFieldMeta fld)
  {
    rec.setNull(fld);
  }
  
  private void setValueToField(SRecordInstance rec, SFieldMeta fld, Object val)
  {
    if (fld instanceof SFieldString) {
      SFieldString fs = (SFieldString) fld;
      rec.setString(fs, (String)val);
    }
    else {
      rec.setObject(fld, val);
    }
  }
  
  /**
   * Get the list of {@link TransportFieldMapping} objects for the given
   * class of transport objects.
   * @param cls
   * @return
   */
  public List<TransportFieldMapping> getFieldMappings(Class<?> cls)
  { 
    return (List<TransportFieldMapping>) mappingsByClass.get(cls);
  }
  
}
