package simpleorm.utils;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import simpleorm.dataset.SFieldScalar;
import simpleorm.dataset.SRecordMeta;
import simpleorm.sessionjdbc.SSessionJdbc;

/**
 * An Object that corresponds to what {@link DatabaseMetaData} methods returns. 
 * @author Karadimas
 *
 */
public class ColumnMetaData
{
  private String name;
  private String typeName;
  private int javaSqlType;
  private int size;
  private int decimalDigits;
  private int nullable;
  private String remarks;
  //String precision;
  //String scale;
  
  public static final List<String> columnDiffs(
      SSessionJdbc s, String schemaPattern, SRecordMeta<?> m)
  throws SQLException
  {
    Connection cxn = s.getJdbcConnection();
    DatabaseMetaData md = cxn.getMetaData();
    String catalog = cxn.getCatalog();    
    String tableName = m.getTableName();    
    ResultSet rs = md.getColumns(catalog, schemaPattern, tableName, null);
    /*
     * Retrieves a description of table columns available in the specified catalog. 

Only column descriptions matching the catalog, schema, table and column name criteria are returned. They are ordered by TABLE_SCHEM, TABLE_NAME, and ORDINAL_POSITION. 

Each column description has the following columns: 

TABLE_CAT String => table catalog (may be null) 
TABLE_SCHEM String => table schema (may be null) 
TABLE_NAME String => table name 
COLUMN_NAME String => column name 
DATA_TYPE int => SQL type from java.sql.Types 
TYPE_NAME String => Data source dependent type name, for a UDT the type name is fully qualified 
COLUMN_SIZE int => column size. For char or date types this is the maximum number of characters, for numeric or decimal types this is precision. 
BUFFER_LENGTH is not used. 
DECIMAL_DIGITS int => the number of fractional digits 
NUM_PREC_RADIX int => Radix (typically either 10 or 2) 
NULLABLE int => is NULL allowed. 
columnNoNulls - might not allow NULL values 
columnNullable - definitely allows NULL values 
columnNullableUnknown - nullability unknown 
REMARKS String => comment describing column (may be null) 
COLUMN_DEF String => default value (may be null) 
SQL_DATA_TYPE int => unused 
SQL_DATETIME_SUB int => unused 
CHAR_OCTET_LENGTH int => for char types the maximum number of bytes in the column 
ORDINAL_POSITION int => index of column in table (starting at 1) 
IS_NULLABLE String => "NO" means column definitely does not allow NULL values; "YES" means the column might allow NULL values. An empty string means nobody knows. 
SCOPE_CATLOG String => catalog of table that is the scope of a reference attribute (null if DATA_TYPE isn't REF) 
SCOPE_SCHEMA String => schema of table that is the scope of a reference attribute (null if the DATA_TYPE isn't REF) 
SCOPE_TABLE String => table name that this the scope of a reference attribure (null if the DATA_TYPE isn't REF) 
SOURCE_DATA_TYPE short => source type of a distinct type or user-generated Ref type, SQL type from java.sql.Types (null if DATA_TYPE isn't DISTINCT or user-generated REF) 
Parameters:
catalog a catalog name; must match the catalog name as it is stored in the database; "" retrieves those without a catalog; null means that the catalog name should not be used to narrow the search
schemaPattern a schema name pattern; must match the schema name as it is stored in the database; "" retrieves those without a schema; null means that the schema name should not be used to narrow the search
tableNamePattern a table name pattern; must match the table name as it is stored in the database 
columnNamePattern a column name pattern; must match the column name as it is stored in the database
Returns:
ResultSet - each row is a column description
Throws:
SQLException - if a database access error occurs
See Also:
     */
    List<ColumnMetaData> colMds = new ArrayList<ColumnMetaData>();
    while (rs.next()) {
      ColumnMetaData cmd = new ColumnMetaData();
      cmd.name = rs.getString("COLUMN_NAME");
      cmd.javaSqlType = rs.getInt("DATA_TYPE");
      cmd.typeName = rs.getString("TYPE_NAME");
      cmd.size = rs.getInt("COLUMN_SIZE");
      cmd.decimalDigits = rs.getInt("DECIMAL_DIGITS");
      cmd.nullable = rs.getInt("NULLABLE");
      cmd.remarks = rs.getString("REMARKS");
      colMds.add(cmd);
    }// while
    rs.close();
    List<String> diffs = new ArrayList<String>();
    SFieldScalar[] scalars = m.getAllScalarFields();
    for (int i = 0; i < scalars.length; i++) {
      SFieldScalar scalar = scalars[i];
      String columnName = scalar.getColumnName();
      int cmdix = findColumnMetaData(colMds, columnName); 
      if (cmdix < 0) {
        diffs.add("1," + columnName + ",,");
      }
      else {
        ColumnMetaData cmd = colMds.get(cmdix);
        if (cmd.size != scalar.getMaxSize()) {
          diffs.add("3," + columnName + "," + scalar.getMaxSize() + "," + cmd.size);
        }
        else if (cmd.javaSqlType != scalar.javaSqlType()) {
          diffs.add("4," + columnName + "," + scalar.javaSqlType() + "," + cmd.javaSqlType);
        }
        else if (!cmd.typeName.equalsIgnoreCase(scalar.defaultSqlDataType())) {
          diffs.add("5," + columnName + "," + scalar.defaultSqlDataType() + "," + cmd.typeName);
        }
      }// else
      scalars[i] = null;
    }
    return null;
  }
  
  public static int findColumnMetaData(List<ColumnMetaData> cmds, String columnName)
  {
    int i = 0;
    for (ColumnMetaData cmd : cmds) {
      if (cmd.name.equals(columnName)) return i;
      i++;
    }// for
    return -1;
  }

  public String getName() {
    return name;
  }

  public void setName(String name) {
    this.name = name;
  }

  public String getTypeName() {
    return typeName;
  }

  public void setTypeName(String typeName) {
    this.typeName = typeName;
  }

  public int getJavaSqlType() {
    return javaSqlType;
  }

  public void setJavaSqlType(int javaSqlType) {
    this.javaSqlType = javaSqlType;
  }

  public int getSize() {
    return size;
  }

  public void setSize(int size) {
    this.size = size;
  }

  public int getDecimalDigits() {
    return decimalDigits;
  }

  public void setDecimalDigits(int decimalDigits) {
    this.decimalDigits = decimalDigits;
  }

  public int getNullable() {
    return nullable;
  }

  public void setNullable(int nullable) {
    this.nullable = nullable;
  }

  public String getRemarks() {
    return remarks;
  }

  public void setRemarks(String remarks) {
    this.remarks = remarks;
  }
  
}
